from unittest import TestCase
from test_tidipy import resolver_composition
from test_tidipy.resolver_composition import Buzz

from tidipy import scan, get_resolver, reset


class TestResolver(TestCase):
    def setUp(self) -> None:
        scan(resolver_composition)
        self.resolver = get_resolver()

    def tearDown(self) -> None:
        reset()

    def test_buzz(self):
        result = self.resolver(Buzz)

        self.assertEqual(result, Buzz('My name: Timber'))

    def test_same_instance(self):
        result_1 = self.resolver(Buzz)
        result_2 = self.resolver(Buzz)

        self.assertEqual(id(result_1), id(result_2))



from typing import Optional

from .resolver import Resolver
from .root_scope_provider import RootScopeProvider
from .scope_context import ScopeContext
from .composer_repository import ComposerRepository
from .scope_type import parse_scope_type


def ensure_scope(
    scope_id: str,
    scope_type: str,
    parent_id: str = 'root',
    context: Optional[dict[str, str]] = None
) -> None:
    parsed_scope_type=parse_scope_type(scope_type)
    parsed_context=ScopeContext(context) if context is not None else ScopeContext.empty()
    root_scope = RootScopeProvider.get()

    existing_scope = root_scope.find_scope(scope_id)

    if existing_scope is None:
        parent_scope = root_scope.find_scope(parent_id)
        parent_scope.add_scope(
            scope_id=scope_id,
            scope_type=parsed_scope_type,
            context=parsed_context
        )
        return

    if not existing_scope.matches(parsed_scope_type, parent_id, parsed_context):
        raise Exception


def get_resolver(scope_id: str = 'root') -> Resolver:
    return RootScopeProvider.get().find_scope(scope_id).resolver()


def clear_scope(scope_id: str) -> None:
    if scope_id == 'root':
        RootScopeProvider.reset()
        return
    return RootScopeProvider.get().remove_scope(scope_id)


def reset() -> None:
    RootScopeProvider.reset()
    ComposerRepository.reset()



from abc import abstractmethod
from typing import Type, TypeVar, Optional

T = TypeVar('T')

class Resolver:
    @abstractmethod
    def __call__(self, dependency_type: Type[T], id: Optional[str] = None) -> T:
        ...



from .composer import Composer


class ComposerRepository:
    _composers: set[Composer] = set()

    @classmethod
    def add_composer(cls, composer: Composer) -> None:
        if composer in cls._composers:
            return
        existing_ids = {composer.id for composer in cls._composers}
        if composer.id in existing_ids:
            raise Exception(f'Duplicate composer with id {composer.id}')

        cls._composers.add(composer)

    @classmethod
    def get_composers(cls) -> set[Composer]:
        return cls._composers

    @classmethod
    def reset(cls) -> None:
        cls._composers = set()




from dataclasses import dataclass

from tidipy import composer, Resolver, auto_compose


@dataclass
class Point:
    x: int
    y: int


@composer
def point() -> Point:
    return Point(x=1, y=2)


@dataclass
class PointWrapper:
    point: Point


@composer
def point_wrapper() -> PointWrapper:
    return PointWrapper(Point(x=3, y=4))


class PointWrapperWrapper:
    def __init__(self, point_wrapper: PointWrapper):
        self.point_wrapper = point_wrapper


class EmptyInit:
    def hey(self) -> str:
        return 'hey'


from typing import Optional

from tidipy.composer_repository import ComposerRepository
from tidipy.scope import Scope
from tidipy.scope_context import ScopeContext
from tidipy.scope_type import RootType


class RootScopeProvider:
    _root_scope: Optional[Scope] = None

    @classmethod
    def get(cls) -> Scope:
        if cls._root_scope is None:
            RootScopeProvider._root_scope = Scope(
                scope_id='root',
                scope_type=RootType(),
                composers=ComposerRepository.get_composers(),
                context=ScopeContext.empty()
            )

        return cls._root_scope

    @classmethod
    def reset(cls) -> None:
        cls._root_scope = None



from __future__ import annotations

from typing import Optional

from .children import Children
from .composer import Composer
from .dependency_bag import DependencyBag
from .resolve_from_dependency_bag import ResolveFromDependencyBag
from .resolver import Resolver
from .scope_context import ScopeContext
from .scope_type import ScopeType, Transient


class Scope:
    def __init__(
        self,
        scope_id: str,
        scope_type: ScopeType,
        composers: set[Composer],
        context: ScopeContext,
        parent: Optional[Scope] = None
    ):
        self._scope_id = scope_id
        self._parent = parent
        self._children: Children[Scope] = Children(self, {})
        self._scope_type = scope_type
        self._composers = composers
        self._context = context
        self._resolver = self._create_resolver()
        self._validate()

    def _create_resolver(self) -> ResolveFromDependencyBag:
        return ResolveFromDependencyBag(
            parent=self._parent.resolver() if self._parent is not None else None,
            dependency_bag=DependencyBag([
                composer
                for composer in self._composers
                if not composer.scope_type.supports_storing() or composer.scope_type == self._scope_type
                and composer.context_filter.is_fulfilled_by(self._context)
            ])
        )

    def _ancestor_has_type(self, scope_type: ScopeType) -> bool:
        if self._parent is None:
            return False

        if self._parent._scope_type == scope_type:
            return True

        return self._parent._ancestor_has_type(scope_type)

    def _validate(self):
        if self._ancestor_has_type(self._scope_type):
            raise Exception('Ancestor already has a scope of this type')
        if self._scope_type == Transient():
            raise Exception('Scopes can not have type transient')

    def get_id(self) -> str:
        return self._scope_id

    def add_scope(
        self,
        scope_id: str,
        scope_type: ScopeType,
        context: ScopeContext
    ):
        self._children.add_child(
            Scope(
                scope_id=scope_id,
                parent=self,
                scope_type=scope_type,
                composers=self._composers,
                context= context.add(self._context)
            )
        )

    def find_scope(self, scope_id: str) -> Optional[Scope]:
        return self._children.find_descendant(scope_id)

    def remove_scope(self, scope_id: str) -> None:
        self._children.remove_descendant(scope_id)

    def matches(self, scope_type: ScopeType, parent_id: Optional[str], context: Optional[ScopeContext]) -> bool:
        if self._scope_type != scope_type:
            return False
        if self._parent is None and parent_id is not None:
            return False
        if self._parent is not None and self._parent.get_id() != parent_id:
            return False
        if context is not None and not context.part_of(self._context):
            return False

        return True

    def resolver(self) -> Resolver:
        return self._resolver



from __future__ import annotations

from dataclasses import dataclass
from typing import Type, Any, Callable

from .context_filter import ContextFilter
from .resolver import Resolver
from .scope_type import ScopeType, RootType

@dataclass(frozen=True)
class Composer:
    id: str
    scope_type: ScopeType
    context_filter: ContextFilter
    factory: Callable[[Resolver], Any]
    dependency_type: Type

    def get_id(self) -> str:
        return self.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __post_init__(self):
        if self.scope_type == RootType() and not self.context_filter.is_empty():
            raise Exception('Dependency with root scope type cannot have context filter')

    def supports_storing(self) -> bool:
        return self.scope_type.supports_storing()



from unittest import TestCase

from test_tidipy import autocompose_composition
from test_tidipy.autocompose_composition import PointWrapper, Point, PointWrapperWrapper, EmptyInit
from tidipy import scan, get_resolver, reset, auto_compose, ensure_scope


class TestAutoCompose(TestCase):
    def setUp(self) -> None:
        scan(autocompose_composition)

    def tearDown(self) -> None:
        reset()

    def test_auto_compose(self):
        auto_compose(PointWrapperWrapper)
        resolver = get_resolver()
        result = resolver(PointWrapperWrapper)

        self.assertEqual(PointWrapper(point=Point(x=3, y=4)), result.point_wrapper)

    def test_auto_compose_with_scope_type(self):
        auto_compose(PointWrapperWrapper, scope_type='tenant')
        resolver = get_resolver()
        ensure_scope('tenant-a', scope_type='tenant')
        tenant_resolver = get_resolver('tenant-a')

        with self.assertRaises(Exception):
            resolver(PointWrapperWrapper)
        self.assertEqual(PointWrapper(point=Point(x=3, y=4)), tenant_resolver(PointWrapperWrapper).point_wrapper)

    def test_auto_compose_with_id(self):
        auto_compose(PointWrapper, id='auto-composed')
        resolver = get_resolver()

        result = resolver(PointWrapper, id='auto-composed')

        self.assertEqual(PointWrapper(point=Point(x=1, y=2)), result)

    def test_duplicate_ids(self):
        auto_compose(str, id='some_id')

        with self.assertRaises(Exception):
            auto_compose(str, id='some_id', scope_type='else')

    def test_empty_init(self):
        auto_compose(EmptyInit)
        resolver = get_resolver()
        result = resolver(EmptyInit)

        self.assertEqual(result.hey(), 'hey')

    def test_cannot_add_conditions_to_root_scoped_dependency(self):
        with self.assertRaises(Exception):
            auto_compose(PointWrapper, environment='test')

    def test_can_add_conditions_to_root_scoped_dependency(self):
        try:
            auto_compose(PointWrapper, scope_type='iets', environment='test')
        except Exception:
            self.fail()



from typing import Protocol, Optional, TypeVar, Generic


class HasId(Protocol):
    def get_id(self) -> str: ...

T = TypeVar('T', bound=HasId)


class Children(Generic[T]):
    def __init__(self, value: T, children: dict[str, T]):
        self._value = value
        self._children = children


    def add_child(
        self,
        child: T
    ):
        self._children[child.get_id()] = child

    def find_descendant(self, node_id: str) -> Optional[T]:
        if node_id == self._value.get_id():
            return self._value

        for child_id, child in self._children.items():
            result = child.find_scope(node_id)
            if result is not None:
                return result

        return None

    def remove_descendant(self, node_id: str) -> None:
        if node_id in self._children:
            self._children.pop(node_id)

        for child_id, child in self._children.items():
            child.remove_scope(node_id)



import inspect
from typing import Type

from .resolver import Resolver


class AutoFactory:
    def __init__(self, dependency_type: Type):
        self._dependency_type = dependency_type

    def __eq__(self, other) -> bool:
        if not isinstance(other, AutoFactory):
            return False

        return self._dependency_type == other._dependency_type

    def __call__(self, resolver: Resolver):
        try:
            sig = inspect.signature(self._dependency_type)
        except ValueError:
            # Built-in types may raise ValueError
            return self._dependency_type()

        params = list(sig.parameters.values())

        # Skip 'self' if it's an instance method (e.g., normal class __init__)
        if params and params[0].name == 'self':
            params = params[1:]

        args = []
        for param in params:
            if param.annotation is inspect.Parameter.empty:
                raise TypeError(f"Cannot resolve untyped parameter: {param.name}")
            dependency = resolver(param.annotation)
            args.append(dependency)

        return self._dependency_type(*args)



from dataclasses import dataclass

from tidipy import composer, Resolver


@composer(id='timber')
def my_name() -> str:
    return 'Timber'


@composer
def other_name() -> str:
    return 'Benji'


@dataclass
class Buzz:
    description: str


@composer
def buzz(resolve: Resolver) -> Buzz:
    name = resolve(str, id='timber')
    return Buzz(description=f'My name: {name}')



from unittest import TestCase

from test_tidipy import context_composition
from test_tidipy.context_composition import StringGenerator, HelloGenerator, TimberGenerator, App
from tidipy import scan, get_resolver, reset, ensure_scope


class TestContext(TestCase):
    def setUp(self) -> None:
        scan(context_composition)

    def tearDown(self) -> None:
        reset()

    def test_production(self):
        ensure_scope(scope_id='app-prod', scope_type='app', context={'environment': 'prod'})
        result = get_resolver('app-prod')(StringGenerator)

        self.assertEqual(result.generate(), 'Timber')

    def test_specific_type(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})
        result = get_resolver('app')(HelloGenerator)

        self.assertEqual(result.generate(), 'Hello')

    def test_resolves_with_additional_values(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test', 'a': 'b'})
        result = get_resolver('app')(HelloGenerator)

        self.assertEqual(result.generate(), 'Hello')

    def test_cannot_resolve_with_multiple_candidates(self):
        with self.assertRaises(Exception):
            get_resolver()(StringGenerator)

    def test_can_resolve_specific(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'prod'})
        result = get_resolver('app')(TimberGenerator)

        self.assertEqual(result.generate(), 'Timber')

    def test_resolve_app(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})
        app = get_resolver('app')(App)

        self.assertEqual(app.generate(), 'Hello')

    def test_resolve_two_apps(self):
        ensure_scope(scope_id='app-test', scope_type='app', context={'environment': 'test'})
        ensure_scope(scope_id='app-prod', scope_type='app', context={'environment': 'prod'})
        prod_app = get_resolver('app-prod')(App)
        test_app = get_resolver('app-test')(App)

        self.assertEqual(test_app.generate(), 'Hello')
        self.assertEqual(prod_app.generate(), 'Timber')

    def test_cannot_change_context(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})

        with self.assertRaises(Exception):
            ensure_scope(scope_id='app', scope_type='app', context={'environment': 'prod'})

    def test_cannot_add_to_context(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})

        with self.assertRaises(Exception):
            ensure_scope(scope_id='app', scope_type='app', context={'a': 'b'})

    def test_cannot_change_empty_context(self):
        ensure_scope(scope_id='app', scope_type='app')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='app', scope_type='app', context={'a': 'b'})

    def test_cannot_change_context_in_child(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})

        with self.assertRaises(Exception):
            ensure_scope('child', scope_type='child', parent_id='app', context={'environment': 'prod'})

    def test_adding_context_to_child_scope_does_not_influence_root(self):
        ensure_scope(scope_id='child', scope_type='child', context={'environment': 'test'})

        root_resolver = get_resolver()

        with self.assertRaises(Exception):
            root_resolver(StringGenerator)

    def test_adding_context_to_child_scope_does_not_influence_root_dependencies(self):
        ensure_scope(scope_id='child', scope_type='child', context={'environment': 'test'})
        resolver = get_resolver('child')

        with self.assertRaises(Exception):
            resolver(str)

    def test_child_scope_inherits_parent_context(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})
        ensure_scope(scope_id='tenant-a', scope_type='tenant', parent_id='app')
        resolver = get_resolver('tenant-a')

        self.assertEqual('Hello', resolver(str))

    def test_child_scope_may_reaffirm_context(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})
        try:
            ensure_scope(scope_id='tenant-a', scope_type='tenant', parent_id='app', context={'environment': 'test'})
        except Exception:
            self.fail()

    def test_ensure_may_without_context(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test'})
        try:
            ensure_scope(scope_id='app', scope_type='app')
        except Exception:
            self.fail()

    def test_ensure_may_with_less_context(self):
        ensure_scope(scope_id='app', scope_type='app', context={'environment': 'test', 'a': 'b'})
        try:
            ensure_scope(scope_id='app', scope_type='app', context={'a': 'b'})
        except Exception:
            self.fail()



from __future__ import annotations

from typing import Optional, Type, Any
from .resolver import Resolver
from .composer import Composer


class DependencyBag:
    def __init__(self, composers: list[Composer]):
        self._composers = composers
        self._dependencies: dict[str, Any] = {}

    def _get_candidates(
        self,
        dependency_type: Type,
        dependency_id: Optional[str]
    ) -> list[Composer]:
        candidates = [
            composer
            for composer in self._composers
            if issubclass(composer.dependency_type, dependency_type)
        ]
        if dependency_id is not None:
            candidates = [composer for composer in candidates if composer.id == dependency_id]

        return candidates

    def find(
        self,
        dependency_type: Type,
        dependency_id: Optional[str],
        resolver: Resolver
    ) -> Optional[Any]:
        candidates = self._get_candidates(dependency_type, dependency_id)

        if len(candidates) == 0:
            return None
        if len(candidates) > 1:
            raise Exception(f'More than 1 candidate for type {dependency_type}')

        composer = next(iter(candidates))
        composer_id = composer.id
        if composer_id in self._dependencies:
            return self._dependencies[composer_id]

        dependency = composer.factory(resolver)

        if composer.supports_storing():
            self._dependencies[composer_id] = dependency

        return dependency



import uuid
from dataclasses import dataclass

from tidipy import composer, Resolver


@dataclass
class Hey:
    age: int


@dataclass
class Animal:
    name: str


@composer
def hey() -> Hey:
    return Hey(17)


@composer(scope_type='tenant')
def animal() -> Animal:
    return Animal(name='Henk')


@dataclass
class User:
    id: str


@composer(scope_type='request')
def user() -> User:
    return User(id='user')


@composer(scope_type='transient')
def a_random_string() -> str:
    return str(uuid.uuid4())



from .composer_decorator import composer
from .resolver import Resolver
from .scan import scan
from .scope_api import ensure_scope, clear_scope, reset, get_resolver
from .auto_compose import auto_compose



from unittest import TestCase

from test_tidipy import duplicate_id_composition
from tidipy import scan, reset


class TestDuplicateId(TestCase):
    def tearDown(self) -> None:
        reset()

    def test_duplicate_id(self):
        with self.assertRaises(Exception):
            scan(duplicate_id_composition)



from dataclasses import dataclass

from tidipy.scope_context import ScopeContext


@dataclass(frozen=True)
class ContextFilterElement:
    key: str
    one_of_values: set[str]

    def is_fulfilled_by(self, context: ScopeContext) -> bool:
        if self.key not in context.values():
            return True

        return context.values()[self.key] in self.one_of_values


@dataclass(frozen=True)
class ContextFilter:
    elements: set[ContextFilterElement]

    def is_fulfilled_by(self, context: ScopeContext) -> bool:
        return all(element.is_fulfilled_by(context) for element in self.elements)

    def is_empty(self) -> bool:
        return len(self.elements) == 0


def parse_context_filter(**kwargs) -> ContextFilter:
    return ContextFilter(
            elements={
                ContextFilterElement(
                    key=key,
                    one_of_values=frozenset(value) if isinstance(value, set) else frozenset({value})  # type: ignore
                )
                for key, value in kwargs.items()
            }
        )



from uuid import uuid4

from fastapi import FastAPI, Depends
from starlette.requests import Request

from tidipy import get_resolver, ensure_scope, clear_scope, auto_compose


class NameSayer:
    def my_name(self) -> str:
        return 'TiDIpy'


auto_compose(NameSayer, scope_type='request')


app = FastAPI()


def di(dependency_type):
    def dependency(request: Request):
        return request.state.resolve(dependency_type)
    return Depends(dependency)


@app.middleware("http")
async def add_request_scope(request: Request, call_next):
    request_id = str(uuid4())
    ensure_scope(request_id, scope_type='request')
    request.state.resolve = get_resolver(request_id)
    try:
        return await call_next(request)
    finally:
        clear_scope(request_id)


@app.get('/say-my-name')
def say_my_name(name_sayer: NameSayer = di(NameSayer)):
    return name_sayer.my_name()



from unittest import TestCase
from test_tidipy import scope_composition

from test_tidipy.scope_composition import Animal, Hey, User

from tidipy import scan, get_resolver, ensure_scope, clear_scope, reset


class TestScope(TestCase):
    def setUp(self) -> None:
        scan(scope_composition)

    def tearDown(self) -> None:
        reset()

    def test_cannot_access_in_root_scope(self):
        resolver = get_resolver()
        with self.assertRaises(Exception):
            resolver(Animal)

    def test_cannot_get_root_scope_with_other_scope(self):
        with self.assertRaises(Exception):
            ensure_scope(scope_id='root', scope_type='transient')

    def test_cannot_get_root_scope_with_parent(self):
        with self.assertRaises(Exception):
            ensure_scope(scope_id='root', scope_type='root', parent_id='other')

    def test_cannot_get_scope_with_nonexisting_parent(self):
        with self.assertRaises(Exception):
            ensure_scope(scope_id='hey', scope_type='hey', parent_id='non-existant')

    def test_create_scope(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        resolver = get_resolver('tenant-a')
        result = resolver(Animal)

        self.assertEqual(Animal(name='Henk'), result)

    def test_ensure_scope(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        resolver = get_resolver('tenant-a')
        result = resolver(Animal)

        self.assertEqual(Animal(name='Henk'), result)

    def test_ensure_twice(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        ensure_scope(scope_id='tenant-a', scope_type='tenant')

        resolver = get_resolver('tenant-a')
        result = resolver(Animal)

        self.assertEqual(Animal(name='Henk'), result)

    def test_ensure_with_other_type(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='tenant-a', scope_type='iets')

    def test_destroy_scope(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        clear_scope(scope_id='tenant-a')

        with self.assertRaises(Exception):
            get_resolver('tenant-a')

    def test_can_access_root_scope_from_child(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        resolver = get_resolver('tenant-a')

        result = resolver(Hey)

        resolved_from_root = get_resolver()(Hey)

        self.assertEqual(Hey(age=17), result)
        self.assertEqual(id(result), id(resolved_from_root))

    def test_can_access_root_scope_after_clearing(self):
        clear_scope('root')
        resolver = get_resolver()

        result = resolver(Hey)

        self.assertEqual(Hey(age=17), result)

    def test_clear_root_scope(self):
        first_object = get_resolver()(Hey)
        clear_scope('root')
        second_object = get_resolver()(Hey)

        self.assertNotEqual(id(first_object), id(second_object))


    def test_cannot_ensure_parent_scope_with_context_after_nesting(self):
        ensure_scope(scope_id='parent', scope_type='parent')
        ensure_scope(scope_id='child', scope_type='child', parent_id='parent')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='parent', scope_type='parent', context={'a': 'b'})

    def test_can_ensure_root_scope_after_clearing(self):
        ensure_scope(scope_id='parent', scope_type='parent')
        ensure_scope(scope_id='child', scope_type='child', parent_id='parent')
        clear_scope('parent')
        try:
            ensure_scope(scope_id='parent', scope_type='parent', context={'a': 'b'})
        except Exception:
            self.fail()

    def test_nested_scopes(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        ensure_scope(scope_id='request-b', scope_type='request', parent_id='tenant-a')
        resolver = get_resolver('request-b')

        user_result = resolver(User)
        hey_result = resolver(Hey)

        self.assertEqual(User(id='user'), user_result)
        self.assertEqual(Hey(age=17), hey_result)

    def test_destroy_child_scopes(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        ensure_scope(scope_id='request-b', scope_type='request', parent_id='tenant-a')
        clear_scope('tenant-a')

        with self.assertRaises(Exception):
            get_resolver('request-b')

    def test_cannot_nest_same_type(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='tenant-b', scope_type='tenant', parent_id='tenant-a')

    def test_illegal_ensures(self):
        with self.assertRaises(Exception):
            ensure_scope(scope_id='tenant-b', scope_type='root')
        with self.assertRaises(Exception):
            ensure_scope(scope_id='root', scope_type='iets')
        with self.assertRaises(Exception):
            ensure_scope(scope_id='root', scope_type='root', parent_id='iets')
        with self.assertRaises(Exception):
            ensure_scope(scope_id='request-a', scope_type='transient')

    def test_cannot_create_child_scope_of_same_type(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='tenant-b', scope_type='tenant', parent_id='tenant-a')

    def test_cannot_nest_scope_of_same_type(self):
        ensure_scope(scope_id='tenant-a', scope_type='tenant')
        ensure_scope(scope_id='request', scope_type='request', parent_id='tenant-a')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='tenant-b', scope_type='tenant', parent_id='request')

    def test_conflicting_scope_types(self):
        ensure_scope(scope_id='x', scope_type='tenant')

        with self.assertRaises(Exception):
            ensure_scope(scope_id='x', scope_type='request')

    def test_transient_dependencies_are_not_stored(self):
        resolver = get_resolver()

        self.assertNotEqual(resolver(str), resolver(str))



from abc import ABC, abstractmethod
from typing import Type
from uuid import uuid4

from fastapi import FastAPI
from starlette.requests import Request

from tidipy import get_resolver, ensure_scope, clear_scope, auto_compose


class NameSayer:
    def my_name(self) -> str:
        return 'TiDIpy'


class Controller(ABC):
    @abstractmethod
    def handle(self):
        ...


class MyController(Controller):
    def __init__(self, name_sayer: NameSayer):
        self._name_sayer = name_sayer

    def handle(self):
        return self._name_sayer.my_name()


auto_compose(NameSayer, scope_type='request')
auto_compose(MyController, scope_type='request')


class FastApiAdapter:
    def __init__(self, controller_type: Type[Controller]):
        self._controller_type = controller_type

    def handle(self, request: Request):
        return request.state.resolve(self._controller_type).handle()


app = FastAPI()


@app.middleware("http")
async def add_request_scope(request: Request, call_next):
    request_id = str(uuid4())
    ensure_scope(request_id, scope_type='request')
    request.state.resolve = get_resolver(request_id)
    try:
        return await call_next(request)
    finally:
        clear_scope(request_id)

app.add_api_route(
    path='/say-my-name',
    methods=['GET'],
    endpoint=FastApiAdapter(MyController).handle
)



import builtins
import inspect
from typing import Type, Optional

from .auto_factory import AutoFactory
from .composer import Composer
from .context_filter import parse_context_filter
from .composer_repository import ComposerRepository
from .scope_type import parse_scope_type


def auto_compose(
    dependency_type: Type,
    *,
    id: Optional[str] = None,
    scope_type: str = 'root',
    **kwargs
):
    if not inspect.isclass(dependency_type):
        raise ValueError
    factory = AutoFactory(dependency_type)
    composer = Composer(
            id=str(builtins.id(dependency_type)) if id is None else id,
            scope_type=parse_scope_type(scope_type),
            context_filter=parse_context_filter(**kwargs),
            dependency_type=dependency_type,
            factory=factory
        )
    ComposerRepository.add_composer(composer)



from __future__ import annotations

import builtins
import inspect
from typing import Callable, get_type_hints, Optional

from .composer import Composer
from .context_filter import parse_context_filter
from .scope_type import parse_scope_type


def composer(
    factory: Optional[Callable] = None,
    *,
    id: Optional[str] = None,
    scope_type: str = 'root',
    **kwargs: str | set[str]
):
    parsed_scope_type = parse_scope_type(scope_type)

    def inner(func: Callable):
        has_parameter = len(inspect.signature(func).parameters) > 0
        return Composer(
            id=str(builtins.id(func)) if id is None else id,
            scope_type=parsed_scope_type,
            context_filter=parse_context_filter(**kwargs),
            factory=func if has_parameter else lambda resolve: func(),
            dependency_type=get_type_hints(func).get('return', object)
        )

    if factory is not None:
        return inner(factory)

    return inner



from __future__ import annotations


class ScopeContext:
    def __init__(self, values: dict[str, str]):
        self._values = values

    @staticmethod
    def empty() -> ScopeContext:
        return ScopeContext({})

    def add(self, context: ScopeContext) -> ScopeContext:
        common_keys = set(context.values().keys()).intersection(set(self._values.keys()))

        if any(self._values[key] != context.values()[key] for key in common_keys):
            raise Exception('Can only add values to scope context, not override values')

        return ScopeContext({**self._values, **context.values()})

    def part_of(self, context: ScopeContext) -> bool:
        return context.add(self) == context

    def __eq__(self, other) -> bool:
        if not isinstance(other, ScopeContext):
            return False

        return self._values == other.values()

    def values(self) -> dict[str, str]:
        return self._values



from abc import ABC, abstractmethod

from tidipy import composer, Resolver


class StringGenerator(ABC):
    @abstractmethod
    def generate(self) -> str:
        pass


class TimberGenerator(StringGenerator):
    def generate(self) -> str:
        return 'Timber'


class HelloGenerator(StringGenerator):
    def generate(self) -> str:
        return 'Hello'


class App:
    def __init__(self, generator: StringGenerator):
        self._generator = generator

    def generate(self) -> str:
        return self._generator.generate()


@composer(scope_type='app', environment='prod')
def timber() -> TimberGenerator:
    return TimberGenerator()


@composer(scope_type='app', environment='test')
def hello() -> HelloGenerator:
    return HelloGenerator()


@composer(scope_type='app')
def app(resolve: Resolver) -> App:
    return App(resolve(StringGenerator))


@composer(scope_type='tenant')
def string(resolve: Resolver) -> str:
    return resolve(App).generate()



from __future__ import annotations

from typing import TypeVar, Type, Optional

from tidipy.dependency_bag import DependencyBag
from tidipy.resolver import Resolver

T = TypeVar('T')


class ResolveFromDependencyBag(Resolver):
    def __init__(self, parent: Optional[ResolveFromDependencyBag], dependency_bag: DependencyBag):
        self._parent = parent
        self._dependency_bag = dependency_bag

    def __call__(self, dependency_type: Type[T], id: Optional[str] = None) -> T:
        result = self._dependency_bag.find(dependency_type, id, self)
        if result is not None:
            return result

        if self._parent is None:
            raise Exception(f'No candidate for type {dependency_type}')

        return self._parent(dependency_type, id)
